""" Pyramid Feature Attention Network for Saliency Detection, CVPR 2019

Model from official source: https://github.com/sairajk/PyTorch-Pyramid-Feature-Attention-Network-for-Saliency-Detection

Little modification from original codebase to adapt code to new pytorch versions
"""

import cv2
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as f
import torch.nn.functional as F
import torchvision.models as models


class SpatialAttention(nn.Module):
    def __init__(self, in_channels, kernel_size=9):
        super(SpatialAttention, self).__init__()

        self.kernel_size = kernel_size
        self.in_channels = in_channels
        pad = (self.kernel_size - 1) // 2  # Padding on one side for stride 1

        self.grp1_conv1k = nn.Conv2d(
            self.in_channels,
            self.in_channels // 2,
            (1, self.kernel_size),
            padding=(0, pad),
        )
        self.grp1_bn1 = nn.BatchNorm2d(self.in_channels // 2)
        self.grp1_convk1 = nn.Conv2d(
            self.in_channels // 2, 1, (self.kernel_size, 1), padding=(pad, 0)
        )
        self.grp1_bn2 = nn.BatchNorm2d(1)

        self.grp2_convk1 = nn.Conv2d(
            self.in_channels,
            self.in_channels // 2,
            (self.kernel_size, 1),
            padding=(pad, 0),
        )
        self.grp2_bn1 = nn.BatchNorm2d(self.in_channels // 2)
        self.grp2_conv1k = nn.Conv2d(
            self.in_channels // 2, 1, (1, self.kernel_size), padding=(0, pad)
        )
        self.grp2_bn2 = nn.BatchNorm2d(1)

    def forward(self, input_):
        # Generate Group 1 Features
        grp1_feats = self.grp1_conv1k(input_)
        grp1_feats = F.relu(self.grp1_bn1(grp1_feats))
        grp1_feats = self.grp1_convk1(grp1_feats)
        grp1_feats = F.relu(self.grp1_bn2(grp1_feats))

        # Generate Group 2 features
        grp2_feats = self.grp2_convk1(input_)
        grp2_feats = F.relu(self.grp2_bn1(grp2_feats))
        grp2_feats = self.grp2_conv1k(grp2_feats)
        grp2_feats = F.relu(self.grp2_bn2(grp2_feats))

        added_feats = torch.sigmoid(torch.add(grp1_feats, grp2_feats))
        added_feats = added_feats.expand_as(input_).clone()

        return added_feats


class ChannelwiseAttention(nn.Module):
    def __init__(self, in_channels):
        super(ChannelwiseAttention, self).__init__()

        self.in_channels = in_channels

        self.linear_1 = nn.Linear(self.in_channels, self.in_channels // 4)
        self.linear_2 = nn.Linear(self.in_channels // 4, self.in_channels)

    def forward(self, input_):
        n_b, n_c, h, w = input_.size()

        feats = F.adaptive_avg_pool2d(input_, (1, 1)).view((n_b, n_c))
        feats = F.relu(self.linear_1(feats))
        feats = torch.sigmoid(self.linear_2(feats))

        # Activity regularizer
        ca_act_reg = torch.mean(feats)

        feats = feats.view((n_b, n_c, 1, 1))
        feats = feats.expand_as(input_).clone()

        return feats, ca_act_reg


vgg_conv1_2 = vgg_conv2_2 = vgg_conv3_3 = vgg_conv4_3 = vgg_conv5_3 = None


def conv_1_2_hook(module, input, output):
    global vgg_conv1_2
    vgg_conv1_2 = output
    return None


def conv_2_2_hook(module, input, output):
    global vgg_conv2_2
    vgg_conv2_2 = output
    return None


def conv_3_3_hook(module, input, output):
    global vgg_conv3_3
    vgg_conv3_3 = output
    return None


def conv_4_3_hook(module, input, output):
    global vgg_conv4_3
    vgg_conv4_3 = output
    return None


def conv_5_3_hook(module, input, output):
    global vgg_conv5_3
    vgg_conv5_3 = output
    return None


class CPFE(nn.Module):
    def __init__(self, feature_layer=None, out_channels=32):
        super(CPFE, self).__init__()

        self.dil_rates = [3, 5, 7]

        # Determine number of in_channels from VGG-16 feature layer
        if feature_layer == "conv5_3":
            self.in_channels = 512
        elif feature_layer == "conv4_3":
            self.in_channels = 512
        elif feature_layer == "conv3_3":
            self.in_channels = 256

        # Define layers
        self.conv_1_1 = nn.Conv2d(
            in_channels=self.in_channels,
            out_channels=out_channels,
            kernel_size=1,
            bias=False,
        )
        self.conv_dil_3 = nn.Conv2d(
            in_channels=self.in_channels,
            out_channels=out_channels,
            kernel_size=3,
            stride=1,
            dilation=self.dil_rates[0],
            padding=self.dil_rates[0],
            bias=False,
        )
        self.conv_dil_5 = nn.Conv2d(
            in_channels=self.in_channels,
            out_channels=out_channels,
            kernel_size=3,
            stride=1,
            dilation=self.dil_rates[1],
            padding=self.dil_rates[1],
            bias=False,
        )
        self.conv_dil_7 = nn.Conv2d(
            in_channels=self.in_channels,
            out_channels=out_channels,
            kernel_size=3,
            stride=1,
            dilation=self.dil_rates[2],
            padding=self.dil_rates[2],
            bias=False,
        )

        self.bn = nn.BatchNorm2d(out_channels * 4)

    def forward(self, input_):
        # Extract features
        conv_1_1_feats = self.conv_1_1(input_)
        conv_dil_3_feats = self.conv_dil_3(input_)
        conv_dil_5_feats = self.conv_dil_5(input_)
        conv_dil_7_feats = self.conv_dil_7(input_)

        # Aggregate features
        concat_feats = torch.cat(
            (conv_1_1_feats, conv_dil_3_feats, conv_dil_5_feats, conv_dil_7_feats),
            dim=1,
        )
        bn_feats = F.relu(self.bn(concat_feats))

        return bn_feats


class PFAN(nn.Module):
    def __init__(self, **kwargs):
        super(PFAN, self).__init__()

        # Load the [partial] VGG-16 model
        self.vgg16 = models.vgg16(weights=models.VGG16_Weights.DEFAULT).features

        # Extract and register intermediate features of VGG-16
        self.vgg16[3].register_forward_hook(conv_1_2_hook)
        self.vgg16[8].register_forward_hook(conv_2_2_hook)
        self.vgg16[15].register_forward_hook(conv_3_3_hook)
        self.vgg16[22].register_forward_hook(conv_4_3_hook)
        self.vgg16[29].register_forward_hook(conv_5_3_hook)

        # Initialize layers for high level (hl) feature (conv3_3, conv4_3, conv5_3) processing
        self.cpfe_conv3_3 = CPFE(feature_layer="conv3_3")
        self.cpfe_conv4_3 = CPFE(feature_layer="conv4_3")
        self.cpfe_conv5_3 = CPFE(feature_layer="conv5_3")

        self.cha_att = ChannelwiseAttention(
            in_channels=384
        )  # in_channels = 3 x (32 x 4)

        self.hl_conv1 = nn.Conv2d(384, 64, (3, 3), padding=1)
        self.hl_bn1 = nn.BatchNorm2d(64)

        # Initialize layers for low level (ll) feature (conv1_2 and conv2_2) processing
        self.ll_conv_1 = nn.Conv2d(64, 64, (3, 3), padding=1)
        self.ll_bn_1 = nn.BatchNorm2d(64)
        self.ll_conv_2 = nn.Conv2d(128, 64, (3, 3), padding=1)
        self.ll_bn_2 = nn.BatchNorm2d(64)
        self.ll_conv_3 = nn.Conv2d(128, 64, (3, 3), padding=1)
        self.ll_bn_3 = nn.BatchNorm2d(64)

        self.spa_att = SpatialAttention(in_channels=64)

        # Initialize layers for fused features (ff) processing
        self.ff_conv_1 = nn.Conv2d(128, 1, (3, 3), padding=1)

    def forward(self, input_):
        global vgg_conv1_2, vgg_conv2_2, vgg_conv3_3, vgg_conv4_3, vgg_conv5_3

        # Pass input_ through vgg16 to generate intermediate features
        self.vgg16(input_)

        # Process high level features
        conv3_cpfe_feats = self.cpfe_conv3_3(vgg_conv3_3)
        conv4_cpfe_feats = self.cpfe_conv4_3(vgg_conv4_3)
        conv5_cpfe_feats = self.cpfe_conv5_3(vgg_conv5_3)

        conv4_cpfe_feats = F.interpolate(
            conv4_cpfe_feats, scale_factor=2, mode="bilinear", align_corners=True
        )
        conv5_cpfe_feats = F.interpolate(
            conv5_cpfe_feats, scale_factor=4, mode="bilinear", align_corners=True
        )

        conv_345_feats = torch.cat(
            (conv3_cpfe_feats, conv4_cpfe_feats, conv5_cpfe_feats), dim=1
        )

        conv_345_ca, ca_act_reg = self.cha_att(conv_345_feats)
        conv_345_feats = torch.mul(conv_345_feats, conv_345_ca)

        conv_345_feats = self.hl_conv1(conv_345_feats)
        conv_345_feats = F.relu(self.hl_bn1(conv_345_feats))
        conv_345_feats = F.interpolate(
            conv_345_feats, scale_factor=4, mode="bilinear", align_corners=True
        )

        # Process low level features
        conv1_feats = self.ll_conv_1(vgg_conv1_2)
        conv1_feats = F.relu(self.ll_bn_1(conv1_feats))
        conv2_feats = self.ll_conv_2(vgg_conv2_2)
        conv2_feats = F.relu(self.ll_bn_2(conv2_feats))

        conv2_feats = F.interpolate(
            conv2_feats, scale_factor=2, mode="bilinear", align_corners=True
        )
        conv_12_feats = torch.cat((conv1_feats, conv2_feats), dim=1)
        conv_12_feats = self.ll_conv_3(conv_12_feats)
        conv_12_feats = F.relu(self.ll_bn_3(conv_12_feats))

        conv_12_sa = self.spa_att(conv_345_feats)
        conv_12_feats = torch.mul(conv_12_feats, conv_12_sa)

        # Fused features
        fused_feats = torch.cat((conv_12_feats, conv_345_feats), dim=1)

        # saldet adaptation
        fused_feats = torch.sigmoid(self.ff_conv_1(fused_feats))
        # fused_feats = self.ff_conv_1(fused_feats)

        return fused_feats



import os
import sys
from typing import Callable, List, Tuple

import torch
from torch.utils.data import Dataset
from tqdm import tqdm

from saldet.io import read_mask, read_rgb


class SaliencyDataset(Dataset):
    """Saliency Detection dataset (with images and masks dir)

    Args:
        root_dir (str): dataset dir with images and masks dirs in it.
        train (bool, optional): if True looks for a train folder, else for a val folder. Defaults to True.
        transform (Callable, optional): data augmentations. Defaults to None.
        max_samples (int, optional): maximum number of samples to load. Defaults to None.
    """

    def __init__(
        self,
        root_dir: str,
        train: bool = True,
        transform: Callable = None,
        max_samples: int = None,
    ) -> None:

        super().__init__()

        root_dir = os.path.join(root_dir, "train" if train else "val")

        assert os.path.exists(root_dir), f"{root_dir} does not exists."
        assert "images" in os.listdir(
            root_dir
        ), f"{root_dir} must have 'images' folder in it."
        assert "masks" in os.listdir(
            root_dir
        ), f"{root_dir} must have 'masks' folder in it."

        self.images_dir = os.path.join(root_dir, "images")
        self.masks_dir = os.path.join(root_dir, "masks")

        self.images, self.masks = self._load_data()

        if max_samples is not None:
            self.images = self.images[:max_samples]
            self.masks = self.masks[:max_samples]

        self.transform = transform

        print(f"#" * 12 + f" {'train' if train else 'val'} dataset recap " + "#" * 12)
        print(f"\t> Root Dir: {root_dir}")
        print(f"\t> Images: {len(self.images)}")
        print(f"\t> Masks: {len(self.masks)}")
        print(f"#" * 40)
        print(f"\n")

    def _load_data(self) -> Tuple[List[str], List[str]]:
        """loads data from images dir and masks dir (checks if an image has its mask)

        Returns:
            Tuple[List[str], List[str]]: images and masks paths
        """

        _images = [f for f in os.listdir(self.images_dir) if not f.startswith(".")]
        _masks = [f for f in os.listdir(self.masks_dir) if not f.startswith(".")]

        images, masks = [], []
        for image in _images:
            image_no_ext = os.path.splitext(image)[0]
            for mask in _masks:
                if os.path.splitext(mask)[0] == image_no_ext:
                    images.append(os.path.join(self.images_dir, image))
                    masks.append(os.path.join(self.masks_dir, mask))
                    _masks.remove(mask)
                    break

        if len(_images) != len(images):
            print(
                f"[WARNING] Found {len(_images)-len(images)} images with no masks. Removing them from dataset."
            )
        return images, masks

    def __getitem__(self, index: int) -> Tuple[torch.Tensor, torch.Tensor]:
        """Return (image, mask) with augmentations applied at given indec

        Args:
            index (int): index

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: (image, mask)
        """

        try:
            image = read_rgb(self.images[index])
        except Exception as e:
            print(f"Found error while loading image: {self.images[index]}")
            print(f"Exception {e}")
            sys.exit(0)
        try:
            mask = read_mask(self.masks[index])

        except Exception as e:
            print(f"Found error while loading mask: {self.masks[index]}")
            print(f"Exception {e}")
            sys.exit(0)

        if self.transform:
            image, mask = self.transform(image=image, mask=mask)

        return image, mask

    def __len__(self) -> int:
        return len(self.images)



import os
from pathlib import Path
from shutil import copy

import pytorch_lightning as pl

from saldet import create_model
from saldet.io import load_config
from saldet.loss import create_criterion
from saldet.lr_scheduler import create_scheduler
from saldet.optimizer import create_optimizer
from saldet.pl import SaliencyPLDataModule, SaliencyPLModel
from saldet.trainer import create_callbacks
from saldet.transform import SaliencyTransform
from saldet.utils import now


def train(
    data_dir: Path,
    config_path: Path,
    output_dir: Path,
    resume_from: Path = None,
    seed: int = 42,
):
    """Train experiment entry point

    Args:
        data_dir (Path): path to data directory.
        config_path (Path): path to yaml configuration file
        output_dir (Path): path to output dir
        resume_from (Path, optional): ckpt from pytorch_lightning to resume training from. Defaults to None.
        seed (int, optional): reproducibility seed. Defaults to 42.
    """
    pl.seed_everything(seed=seed, workers=True)
    config = load_config(config_path)
    output_dir = os.path.join(output_dir, now())

    print(f"> Output dir will be {output_dir}")
    os.makedirs(output_dir)
    copy(config_path, os.path.join(output_dir, "config.yaml"))

    # datamodule
    datamodule = SaliencyPLDataModule(
        root_dir=data_dir,
        train_transform=SaliencyTransform(train=True, **config["transform"]),
        val_transform=SaliencyTransform(train=False, **config["transform"]),
        **config["datamodule"],
    )

    model = create_model(**config["model"])
    criterion = create_criterion(**config["loss"])
    optimizer = create_optimizer(params=model.parameters(), **config["optimizer"])
    lr_scheduler = create_scheduler(optimizer=optimizer, **config["lr_scheduler"])

    pl_model = SaliencyPLModel(
        model=model, criterion=criterion, optimizer=optimizer, lr_scheduler=lr_scheduler
    )

    callbacks = create_callbacks(output_dir=output_dir, **config["callbacks"])

    if resume_from is not None:
        print(f"> Resuming training from {resume_from}.")

    trainer = pl.Trainer(logger=False, callbacks=callbacks, **config["trainer"])

    # fit
    print(f"Launching training...")
    trainer.fit(model=pl_model, datamodule=datamodule, ckpt_path=resume_from)



from typing import Tuple, Union

import albumentations as A
import numpy as np
import PIL
import torch
from PIL import Image


class SaliencyTransform:
    """Saliency based transformation

    Args:
        train (bool): train/val mode.
        input_size (Union[int, list, tuple]): image input size.
        interpolation (int, optional): resize interpolation. Defaults to 3.
        mean (list, optional): normalization mean. Defaults to [0.485, 0.456, 0.406].
        std (list, optional): normalization std. Defaults to [0.229, 0.224, 0.225].
        random_crop_p (float, optional): random crop probability. Defaults to 0.1.
        h_flip_p (float, optional): horizontal flip probability. Defaults to 0.1.
        v_flip_p (float, optional): vertical flip probability. Defaults to 0.1.
    """

    def __init__(
        self,
        train: bool,
        input_size: Union[int, list, tuple],
        interpolation: int = 3,
        mean: list = [0.485, 0.456, 0.406],
        std: list = [0.229, 0.224, 0.225],
        random_crop_p: float = 0.1,
        h_flip_p: float = 0.1,
        v_flip_p: float = 0.1,
    ) -> None:

        if isinstance(input_size, tuple) or isinstance(input_size, list):
            height = input_size[0]
            width = input_size[1]
        else:
            height = input_size
            width = input_size

        if train:
            self.transform = A.Compose(
                [
                    A.Resize(
                        height=height,
                        width=width,
                        interpolation=interpolation,
                        always_apply=True,
                    ),
                    A.RandomResizedCrop(height=height, width=width, p=random_crop_p),
                    A.HorizontalFlip(p=h_flip_p),
                    A.VerticalFlip(p=v_flip_p),
                    A.Normalize(mean=mean, std=std),
                ]
            )
        else:
            self.transform = A.Compose(
                [
                    A.Resize(
                        height=height,
                        width=width,
                        interpolation=interpolation,
                        always_apply=True,
                    ),
                    A.Normalize(mean=mean, std=std),
                ]
            )

    def __call__(
        self,
        image: Union[np.array, PIL.Image.Image],
        mask: Union[np.array, PIL.Image.Image] = None,
    ) -> Tuple[np.array, np.array]:
        """Apply augmentations

        Args:
            img (Union[np.array, PIL.Image.Image]): input image
            mask (Union[np.array, PIL.Image.Image], optional): input mask. Defaults to None.

        Returns:
            Tuple[np.array, np.array, np.array]: vanilla img (resize + normalize), view 1, view 2
        """

        if isinstance(image, Image.Image):
            image = np.array(image)

        if isinstance(mask, Image.Image):
            image = np.array(image)

        if mask is not None:
            sample = self.transform(image=image, mask=mask)
            image, mask = sample["image"], sample["mask"]
        else:
            image = self.transform(image=image)["image"]
            mask = None

        image = torch.from_numpy(image.transpose(2, 0, 1))
        if mask is not None:
            mask = torch.from_numpy(mask).long()
            mask = mask.unsqueeze(dim=0)

        return image, mask



import torch
import torch.nn as nn
import torch.nn.functional as F

from saldet.utils.device import device


class EdgeSaliencyLoss(nn.Module):
    def __init__(self, alpha_sal=0.7):
        super(EdgeSaliencyLoss, self).__init__()

        self.alpha_sal = alpha_sal

        self.laplacian_kernel = torch.tensor(
            [[-1.0, -1.0, -1.0], [-1.0, 8.0, -1.0], [-1.0, -1.0, -1.0]],
            dtype=torch.float,
            requires_grad=False,
        )
        self.laplacian_kernel = self.laplacian_kernel.view(
            (1, 1, 3, 3)
        )  # Shape format of weight for convolution
        self.laplacian_kernel = self.laplacian_kernel.to(device())

    @staticmethod
    def weighted_bce(input_, target, weight_0=1.0, weight_1=1.0, eps=1e-15):
        wbce_loss = -weight_1 * target * torch.log(input_ + eps) - weight_0 * (
            1 - target
        ) * torch.log(1 - input_ + eps)
        return torch.mean(wbce_loss)

    def forward(self, y_pred, y_gt):
        # Generate edge mapsx
        y_gt_edges = F.relu(
            torch.tanh(F.conv2d(y_gt.float(), self.laplacian_kernel, padding=(1, 1)))
        )
        y_pred_edges = F.relu(
            torch.tanh(F.conv2d(y_pred, self.laplacian_kernel, padding=(1, 1)))
        )

        # sal_loss = F.binary_cross_entropy(input=y_pred, target=y_gt)
        sal_loss = self.weighted_bce(
            input_=y_pred, target=y_gt, weight_0=1.0, weight_1=1.12
        )
        edge_loss = F.binary_cross_entropy(input=y_pred_edges, target=y_gt_edges)

        total_loss = self.alpha_sal * sal_loss + (1 - self.alpha_sal) * edge_loss
        return total_loss



from typing import Any, List, Tuple, Union

import torch
import torch.nn as nn
from torch import Tensor


class MultiBCELoss(nn.Module):
    def __init__(
        self, weight: Tensor = None, size_average: bool = True, reduce: Any = None
    ) -> None:
        super().__init__()
        self.bce_loss = nn.BCEWithLogitsLoss(
            weight=weight, size_average=size_average, reduce=reduce
        )

    def forward(
        self, preds: Union[List[Tensor], Tensor], target: Tensor
    ) -> Tuple[Tensor, Tensor]:
        """Forward method

        Args:
            preds (Union[List[Tensor], Tensor]): single pred or list of preds
            target (Tensor): target mask

        Returns:
            Tuple[Tensor, Tensor]: total loss
        """

        if not hasattr(preds, "__iter__"):
            return self.bce_loss(preds, target.float())
        else:
            loss = None
            for pred in preds:
                if loss is None:
                    loss = self.bce_loss(pred, target.float())
                else:
                    loss += self.bce_loss(pred, target.float())
            return loss



import torch


def device() -> str:
    """Return device type

    Returns:
        str: device type
    """
    if torch.cuda.is_available():
        return "cuda"
    if torch.has_mps:
        return "mps"
    return "cpu"



import os
from pathlib import Path

import torch
import torch.nn.functional as F
from PIL import Image
from torch.utils.data import DataLoader
from tqdm import tqdm

from saldet import ops
from saldet.dataset import InferenceDataset
from saldet.io import load_config
from saldet.models import load_checkpoint
from saldet.transform import SaliencyTransform
from saldet.utils import device


def inference(
    images_dir: Path,
    config_path: Path,
    ckpt: Path,
    output_dir: Path,
    sigmoid: bool,
    resize_to_original: bool = True,
    threshold: float = None,
):
    """Inference entry point - create saliency maps and save them all in an output dir

    Args:
        images_dir (Path): path to folder with images
        config_path (Path): path to saldet configuration file
        ckpt (Path): path to model ckpt (.pth or .ckpt from pytorch_lightning)
        output_dir (Path): path to output dir where saliency maps will be saved
        sigmoid (bool): if True, applies sigmoid to predicted mask.
        resize_to_original (bool, optional): if True, resize_to_original saliency maps. Defaults to True.
        threshold (float, optional): if True, applies a threshold to saliency maps to generate binary saliency maps. Defaults to None.
    """
    config = load_config(path=config_path)

    model = load_checkpoint(ckpt=ckpt, model_name=config["model"]["model_name"])

    os.makedirs(output_dir, exist_ok=True)
    print(f"> Loading images dataset")
    dataset = InferenceDataset(
        root_dir=images_dir,
        transform=SaliencyTransform(train=False, **config["transform"]),
    )
    data_loader = DataLoader(dataset=dataset, batch_size=1)

    print(f"> Running inference on images and saving saliency maps at {output_dir}")
    model.to(device())
    model.eval()
    with torch.no_grad():
        for _, batch in tqdm(
            enumerate(data_loader), total=len(data_loader), desc="Inference"
        ):
            x, image_path = batch
            x = x.to(device())
            preds = model(x)
            if isinstance(preds, tuple) or isinstance(preds, list):
                preds = preds[0]
            if sigmoid:
                preds = torch.sigmoid(preds)
            if resize_to_original:
                w, h = Image.open(image_path[0]).size
                preds = F.interpolate(preds, size=(h, w), mode="bilinear")
            for pred, image_path in zip(preds, image_path):
                image_name = os.path.basename(image_path).split(".")[0]
                file_path = os.path.join(output_dir, f"{image_name}.png")
                if threshold is not None:
                    pred[pred >= threshold] = 255
                    pred[pred < threshold] = 0
                else:
                    pred *= 255

                pred = pred.squeeze().cpu().numpy()
                pred = Image.fromarray(pred).convert("L")

                pred.save(file_path)



import os
from typing import List

import pytorch_lightning
from pytorch_lightning.callbacks import EarlyStopping, ModelCheckpoint


def create_callbacks(
    output_dir: str,
    filename: str = "epoch={epoch}-step={step}-val_loss={loss/val:.3f}",
    monitor: str = "loss/val",
    mode: str = "min",
    save_top_k: int = 5,
    patience: int = 10,
) -> List[pytorch_lightning.Callback]:
    """Return list of callbacks for Trainer

    Args:
        output_dir (str): output dir
        filename (str, optional): checkpoint filename. Defaults to epoch={epoch}-step={step}-val_loss={loss/val:.3f}-val_iou={IoU/all/val:.3f}.
        monitor (str, optional): metric to monitor. Defaults to "loss/val".
        mode (str, optional): monitor mode. Defaults to "min".
        save_top_k (int, optional): number of best models to save based on monitor. Defaults to 5.
        patience (int, optional): early stopping patience. Defaults to 10.

    Returns:
        List[pytorch_lightning.Callback]: list of Callbacks.
    """
    callbacks = []
    callbacks.append(
        ModelCheckpoint(
            dirpath=os.path.join(output_dir, "checkpoints"),
            filename=filename,
            monitor=monitor,
            verbose=True,
            mode=mode,
            save_top_k=save_top_k,
        )
    )

    callbacks.append(
        EarlyStopping(
            monitor=monitor,
            min_delta=0.0,
            patience=patience,
            verbose=True,
            mode="min",
            check_finite=True,
            stopping_threshold=None,
            divergence_threshold=None,
            check_on_train_epoch_end=None,
        )
    )

    return callbacks



from typing import Callable, Dict, List, Optional, Union

import pytorch_lightning as pl
from torch.utils.data import DataLoader

from saldet.dataset import SaliencyDataset


class SaliencyPLDataModule(pl.LightningDataModule):
    """PyTorch Lightning Saliency DataModule

    Args:
        root_dir (str): dataset root dir
        batch_size (int): batch size
        train_transform (Callable): train data augmentations
        val_transform (Callable): val data augmentations
        shuffle (bool, optional): whether to shuffle dataset . Defaults to True.
        num_workers (int, optional): data laoder num workers. Defaults to 1.
        pin_memory (bool, optional): data loader oin memory. Defaults to False.
        drop_last (bool, optional): data loader drop last. Defaults to False.
        max_samples (int, optional): maximum number of samples for each dataset (only for test). Defaults to None.
    """

    def __init__(
        self,
        root_dir: str,
        batch_size: int,
        train_transform: Callable,
        val_transform: Callable,
        shuffle: bool = True,
        num_workers: int = 1,
        pin_memory: bool = False,
        drop_last: bool = False,
        max_samples: int = None,
    ):

        super().__init__()
        self.root_dir = root_dir
        self.batch_size = batch_size
        self.train_transform = train_transform
        self.val_transform = val_transform
        self.shuffle = shuffle
        self.num_workers = num_workers
        self.pin_memory = pin_memory
        self.drop_last = drop_last
        self.max_samples = max_samples

    def prepare_data(self) -> None:  # type: ignore
        pass

    def setup(self, stage: Optional[str] = None) -> None:
        """loads the data

        Args:
            stage (Optional[str], optional): pipeline stage (fit, validate, test, predict). Defaults to None.
        """
        if stage == "fit" or stage is None:
            self.train_dataset = SaliencyDataset(
                root_dir=self.root_dir,
                train=True,
                transform=self.train_transform,
                max_samples=self.max_samples,
            )

            self.val_dataset = SaliencyDataset(
                root_dir=self.root_dir,
                train=False,
                transform=self.val_transform,
                max_samples=self.max_samples,
            )

        if stage == "test" or stage is None:
            self.test_dataset = SaliencyDataset(
                root_dir=self.root_dir,
                train=False,
                transform=self.val_transform,
                max_samples=self.max_samples,
            )

    def train_dataloader(
        self,
    ) -> Union[DataLoader, List[DataLoader], Dict[str, DataLoader]]:
        return DataLoader(
            dataset=self.train_dataset,
            batch_size=self.batch_size,
            shuffle=self.shuffle,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            drop_last=self.drop_last,
        )

    def val_dataloader(self) -> Union[DataLoader, List[DataLoader]]:
        return DataLoader(
            dataset=self.val_dataset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            drop_last=False,
        )

    def test_dataloader(self) -> Union[DataLoader, List[DataLoader]]:
        return DataLoader(
            dataset=self.val_dataset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            drop_last=False,
        )



import torch
import torch.nn as nn
from torch.nn import functional as F


class BCEIoULoss(nn.Module):
    def __init__(self) -> None:
        super().__init__()

    def forward(self, pred: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:

        size = pred.size()[2:]
        mask = F.interpolate(mask, size=size, mode="bilinear")
        wbce = F.binary_cross_entropy_with_logits(pred, mask.float())
        pred = torch.sigmoid(pred)
        inter = (pred * mask).sum(dim=(2, 3))
        union = (pred + mask).sum(dim=(2, 3))
        wiou = 1 - (inter + 1) / (union - inter + 1)
        return (wbce + wiou).mean()



import torch.nn as nn
from timm import create_model
from torch import functional as F


class ResNet(nn.Module):
    def __init__(self, model_name: str = "resnet18", pretrained: bool = True):
        super().__init__()
        self.model = create_model(model_name, pretrained=pretrained, num_classes=0)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):

        out1 = self.relu(self.model.bn1(self.model.conv1(x)))
        out1 = self.model.maxpool(out1)
        out2 = self.model.layer1(out1)
        out3 = self.model.layer2(out2)
        out4 = self.model.layer3(out3)
        out5 = self.model.layer4(out4)

        return out2, out3, out4, out5



import torch
import torch.nn.functional as F

from saldet.utils import device


def flat(mask: torch.Tensor) -> torch.tensor:
    """Flat a mask

    Args:
        mask (torch.Tensor): input mask

    Returns:
        torch.tensor: flattened mask
    """
    batch_size = mask.shape[0]
    h = 28
    mask = F.interpolate(mask, size=(int(h), int(h)), mode="bilinear")
    x = mask.view(batch_size, 1, -1).permute(0, 2, 1)
    # print(x.shape)  b 28*28 1
    if device() == "mps":
        x = x.float()
    g = x @ x.transpose(-2, -1)  # b 28*28 28*28
    g = g.unsqueeze(1)  # b 1 28*28 28*28
    return g


def normalize(x: torch.Tensor) -> torch.Tensor:
    """Normalize a tensor based on max and min of tensor

    Args:
        x (torch.Tensor): input tensor

    Returns:
        torch.Tensor: normalized tensor
    """
    t_max = torch.max(x)
    t_min = torch.min(x)
    out = (x - t_min) / (t_max - t_min)
    return out



from typing import Dict

import torch
import torch.nn as nn
import torch.nn.functional as F
from timm import create_model

from .resnet import ResNet
from .swin import SwinTransformer

Act = nn.ReLU

__all__ = ["PGNet"]


def state_dict_from_timm() -> Dict:
    """loads swin_base_patch4_window7_224_in22k timm weights for custom Swin layers

    Args:

    Returns:
        Dict: pretrained swin_base_patch4_window7_224 dict
    """

    timm_model = create_model(
        "swin_base_patch4_window7_224", pretrained=True, num_classes=0
    )

    swin = SwinTransformer(224)

    pretrained_state_dict = {}

    timm_model_state_dict = timm_model.state_dict()
    for k, _ in swin.state_dict().items():
        pretrained_state_dict[k] = timm_model_state_dict[k]

    return pretrained_state_dict


def weight_init(module):
    for n, m in module.named_children():
        if isinstance(m, nn.Conv2d):
            nn.init.kaiming_normal_(m.weight, mode="fan_in", nonlinearity="relu")
            if m.bias is not None:
                nn.init.zeros_(m.bias)
        elif isinstance(m, (nn.BatchNorm2d, nn.InstanceNorm2d, nn.BatchNorm1d)):
            nn.init.ones_(m.weight)
            if m.bias is not None:
                nn.init.zeros_(m.bias)
        elif isinstance(m, nn.Linear):
            nn.init.kaiming_normal_(m.weight, mode="fan_in", nonlinearity="relu")
            if m.bias is not None:
                nn.init.zeros_(m.bias)
        elif isinstance(m, nn.Sequential):
            weight_init(m)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, (nn.ReLU, Act, nn.AdaptiveAvgPool2d, nn.Softmax)):
            pass
        else:
            m.initialize()


class Grafting(nn.Module):
    def __init__(self, dim, num_heads=8, qkv_bias=True, qk_scale=None):
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = qk_scale or head_dim**-0.5
        self.k = nn.Linear(dim, dim, bias=qkv_bias)
        self.qv = nn.Linear(dim, dim * 2, bias=qkv_bias)
        self.proj = nn.Linear(dim, dim)
        self.act = nn.ReLU(inplace=True)
        self.conv = nn.Conv2d(8, 8, kernel_size=3, stride=1, padding=1)
        self.lnx = nn.LayerNorm(64)
        self.lny = nn.LayerNorm(64)
        self.bn = nn.BatchNorm2d(8)
        self.conv2 = nn.Sequential(
            nn.Conv2d(64, 64, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(64),
            nn.ReLU(inplace=True),
            nn.Conv2d(64, 64, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(64),
            nn.ReLU(inplace=True),
        )

    def forward(self, x, y):
        batch_size = x.shape[0]
        chanel = x.shape[1]
        sc = x
        x = x.view(batch_size, chanel, -1).permute(0, 2, 1)
        sc1 = x
        x = self.lnx(x)
        y = y.view(batch_size, chanel, -1).permute(0, 2, 1)
        y = self.lny(y)

        B, N, C = x.shape
        y_k = (
            self.k(y)
            .reshape(B, N, 1, self.num_heads, C // self.num_heads)
            .permute(2, 0, 3, 1, 4)
        )
        x_qv = (
            self.qv(x)
            .reshape(B, N, 2, self.num_heads, C // self.num_heads)
            .permute(2, 0, 3, 1, 4)
        )
        x_q, x_v = x_qv[0], x_qv[1]
        y_k = y_k[0]
        attn = (x_q @ y_k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)
        x = (attn @ x_v).transpose(1, 2).reshape(B, N, C)

        x = self.proj(x)
        x = x + sc1

        x = x.permute(0, 2, 1)
        x = x.view(batch_size, chanel, *sc.size()[2:])
        x = self.conv2(x) + x
        return x, self.act(self.bn(self.conv(attn + attn.transpose(-1, -2))))

    def initialize(self):
        weight_init(self)


class DB1(nn.Module):
    def __init__(self, inplanes, outplanes):
        super(DB1, self).__init__()
        self.squeeze1 = nn.Sequential(
            nn.Conv2d(inplanes, outplanes, kernel_size=1, stride=1, padding=0),
            nn.BatchNorm2d(64),
            nn.ReLU(inplace=True),
        )
        self.squeeze2 = nn.Sequential(
            nn.Conv2d(64, 64, kernel_size=3, stride=1, dilation=2, padding=2),
            nn.BatchNorm2d(64),
            nn.ReLU(inplace=True),
        )

    def forward(self, x):
        z = self.squeeze2(self.squeeze1(x))
        return z, z

    def initialize(self):
        weight_init(self)


class DB2(nn.Module):
    def __init__(self, inplanes, outplanes):
        super(DB2, self).__init__()
        self.short_cut = nn.Conv2d(
            outplanes, outplanes, kernel_size=1, stride=1, padding=0
        )
        self.conv = nn.Sequential(
            nn.Conv2d(
                inplanes + outplanes, outplanes, kernel_size=3, stride=1, padding=1
            ),
            nn.BatchNorm2d(outplanes),
            nn.ReLU(inplace=True),
            nn.Conv2d(outplanes, outplanes, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(outplanes),
            nn.ReLU(inplace=True),
        )
        self.conv2 = nn.Sequential(
            nn.Conv2d(outplanes, outplanes, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(outplanes),
            nn.ReLU(inplace=True),
            nn.Conv2d(outplanes, outplanes, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(outplanes),
            nn.ReLU(inplace=True),
        )

    def forward(self, x, z):
        z = F.interpolate(z, size=x.size()[2:], mode="bilinear", align_corners=True)
        p = self.conv(torch.cat((x, z), 1))
        sc = self.short_cut(z)
        p = p + sc
        p2 = self.conv2(p)
        p = p + p2
        return p, p

    def initialize(self):
        weight_init(self)


class DB3(nn.Module):
    def __init__(self) -> None:
        super(DB3, self).__init__()

        self.db2 = DB2(64, 64)

        self.conv3x3 = nn.Sequential(
            nn.Conv2d(64, 64, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(64),
            nn.ReLU(inplace=True),
        )
        self.sqz_r4 = nn.Sequential(
            nn.Conv2d(256, 64, kernel_size=3, stride=1, dilation=1, padding=1),
            nn.BatchNorm2d(64),
            nn.ReLU(inplace=True),
        )

        self.sqz_s1 = nn.Sequential(
            nn.Conv2d(128, 64, kernel_size=3, stride=1, dilation=1, padding=1),
            nn.BatchNorm2d(64),
            nn.ReLU(inplace=True),
        )

    def forward(self, s, r, up):
        up = F.interpolate(up, size=s.size()[2:], mode="bilinear", align_corners=True)
        s = self.sqz_s1(s)
        r = self.sqz_r4(r)
        sr = self.conv3x3(s + r)
        out, _ = self.db2(sr, up)
        return out, out

    def initialize(self):
        weight_init(self)


class Decoder(nn.Module):
    def __init__(self) -> None:
        super(Decoder, self).__init__()
        self.sqz_s2 = nn.Sequential(
            nn.Conv2d(256, 64, kernel_size=3, stride=1, dilation=1, padding=1),
            nn.BatchNorm2d(64),
            nn.ReLU(inplace=True),
        )
        self.sqz_r5 = nn.Sequential(
            nn.Conv2d(512, 64, kernel_size=3, stride=1, dilation=1, padding=1),
            nn.BatchNorm2d(64),
            nn.ReLU(inplace=True),
        )

        self.GF = Grafting(64, num_heads=8)
        self.d1 = DB1(512, 64)
        self.d2 = DB2(512, 64)
        self.d3 = DB2(64, 64)
        self.d4 = DB3()
        self.d5 = DB2(128, 64)
        self.d6 = DB2(64, 64)

    def forward(self, s1, s2, s3, s4, r2, r3, r4, r5):
        r5 = F.interpolate(r5, size=s2.size()[2:], mode="bilinear", align_corners=True)
        s1 = F.interpolate(s1, size=r4.size()[2:], mode="bilinear", align_corners=True)

        s4_, _ = self.d1(s4)
        s3_, _ = self.d2(s3, s4_)

        s2_ = self.sqz_s2(s2)
        r5_ = self.sqz_r5(r5)
        graft_feature_r5, cam = self.GF(r5_, s2_)

        graft_feature_r5_, _ = self.d3(graft_feature_r5, s3_)

        graft_feature_r4, _ = self.d4(s1, r4, graft_feature_r5_)

        r3_, _ = self.d5(r3, graft_feature_r4)

        r2_, _ = self.d6(r2, r3_)

        return r2_, cam, r5_, s2_

    def initialize(self):
        weight_init(self)


class PGNet(nn.Module):
    def __init__(self, input_size: int = 224, pretrained: bool = True, cfg=None):
        super(PGNet, self).__init__()
        self.cfg = cfg
        self.decoder = Decoder()
        self.linear1 = nn.Conv2d(64, 1, kernel_size=3, stride=1, padding=1)
        self.linear2 = nn.Conv2d(64, 1, kernel_size=3, stride=1, padding=1)
        self.linear3 = nn.Conv2d(64, 1, kernel_size=3, stride=1, padding=1)
        self.conv = nn.Conv2d(8, 1, kernel_size=3, stride=1, padding=1)

        if self.cfg is None or self.cfg.snapshot is None:
            weight_init(self)

        self.resnet = ResNet(model_name="resnet18", pretrained=pretrained)
        self.swin = SwinTransformer(input_size)

        if pretrained:
            self.swin.load_state_dict(state_dict_from_timm())

        # self.swin = SwinTransformer(224)
        # self.swin.load_state_dict(torch.load('../pre/swin224.pth')['model'],strict=False)
        # self.resnet.load_state_dict(torch.load('../pre/resnet18.pth'),strict=False)

        # if self.cfg is not None and self.cfg.snapshot:
        #     print('load checkpoint')
        #     pretrain=torch.load(self.cfg.snapshot)
        #     new_state_dict = {}
        #     for k,v in pretrain.items():
        #         new_state_dict[k[7:]] = v
        #     self.load_state_dict(new_state_dict, strict=False)

    def forward(self, x, shape=None, mask=None):
        shape = x.size()[2:] if shape is None else shape
        y = F.interpolate(x, size=(224, 224), mode="bilinear", align_corners=True)

        r2, r3, r4, r5 = self.resnet(x)
        s1, s2, s3, s4 = self.swin(y)
        r2_, attmap, r5_, s2_ = self.decoder(s1, s2, s3, s4, r2, r3, r4, r5)

        pred1 = F.interpolate(self.linear1(r2_), size=shape, mode="bilinear")
        wr = F.interpolate(self.linear2(r5_), size=(28, 28), mode="bilinear")
        ws = F.interpolate(self.linear3(s2_), size=(28, 28), mode="bilinear")

        return pred1, wr, ws, self.conv(attmap)



import sys
from typing import Dict

import yaml


def load_config(path: str) -> Dict:
    """loads a single yml file

    Args:
        path (str): path to yml file

    Returns:
        Dict: dict with configurations (if fail, params is None)
    """
    try:
        with open(path, "r") as stream:
            params = yaml.safe_load(stream)
    except:
        params = None
    return params



import os
from typing import Callable, List, Tuple

import torch
from torch.utils.data import Dataset
from tqdm import tqdm

from saldet.io import read_mask, read_rgb


class InferenceDataset(Dataset):
    """Inference Dataset (a folder with images)

    Args:
        root_dir (str): root data dir (must be with train and val folders)
        transform (Callable, optional): set of data transformations. Defaults to None.

    Raises:
        FileNotFoundError: if something is found erroneous in the dataset
    """

    EXTENSIONS = (
        "jpg",
        "jpeg",
        "png",
        "ppm",
        "bmp",
        "pgm",
        "tif",
        "tiff",
        "webp",
    )

    def __init__(
        self,
        root_dir: str,
        transform: Callable = None,
    ) -> None:

        super().__init__()
        # checking structure
        try:
            self._sanity_check(root_dir=root_dir)
        except Exception as e:
            raise e

        self.data_dir = root_dir
        self.images = [
            os.path.join(self.data_dir, f)
            for f in os.listdir(self.data_dir)
            if f.split(".")[-1].lower() in self.EXTENSIONS
        ]
        self.transform = transform

    def _sanity_check(
        self,
        root_dir: str,
    ):
        """Checks dataset structure

        Args:
            root_dir (str): data directory

        Raises:
            FileNotFoundError: if the data folder is not right based on the structure in class_map
            FileExistsError: if some label does not have images in its folder

        """
        if not (os.path.exists(root_dir)):
            raise FileNotFoundError(f"Folder {root_dir} does not exist")
        images = [
            f
            for f in os.listdir(root_dir)
            if f.split(".")[-1].lower() in self.EXTENSIONS
        ]
        if len(images) == 0:
            raise FileExistsError(f"Folder {root_dir} does not have images.")

        print(f"> Inference dataset sanity check OK")

    def __getitem__(self, index) -> Tuple[torch.Tensor, str]:
        """Return a tuple (image, image path)

        Args:
            index (int): index of dataset

        Returns:
            Tuple[torch.Tensor, str]: image, image path
        """

        image_path = self.images[index]
        image_name = image_path.split(os.sep)[-1]
        image = read_rgb(image_path)

        if self.transform is not None:
            image, _ = self.transform(image, None)

        return image, image_path

    def __len__(self):
        return len(self.images)



import torch.nn as nn

from saldet.loss.edge_saliency_loss import EdgeSaliencyLoss
from saldet.loss.multi_bce import MultiBCELoss
from saldet.loss.pg import PGLoss

__all__ = ["create_criterion"]

FACTORY = {
    "multibce": MultiBCELoss,
    "bce": nn.BCELoss,
    "pg": PGLoss,
    "pfan": EdgeSaliencyLoss,
}


def create_criterion(criterion: str, **kwargs) -> nn.Module:
    return FACTORY[criterion](**kwargs)



import torch
import torch.nn as nn
import torch.utils.checkpoint as checkpoint
from timm.models.layers import DropPath, to_2tuple, trunc_normal_


class Mlp(nn.Module):
    def __init__(
        self,
        in_features,
        hidden_features=None,
        out_features=None,
        act_layer=nn.GELU,
        drop=0.0,
    ):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x


def window_partition(x, window_size):
    """
    Args:
        x: (B, H, W, C)
        window_size (int): window size

    Returns:
        windows: (num_windows*B, window_size, window_size, C)
    """
    B, H, W, C = x.shape
    x = x.view(B, H // window_size, window_size, W // window_size, window_size, C)
    windows = (
        x.permute(0, 1, 3, 2, 4, 5).contiguous().view(-1, window_size, window_size, C)
    )
    return windows


def window_reverse(windows, window_size, H, W):
    """
    Args:
        windows: (num_windows*B, window_size, window_size, C)
        window_size (int): Window size
        H (int): Height of image
        W (int): Width of image

    Returns:
        x: (B, H, W, C)
    """
    B = int(windows.shape[0] / (H * W / window_size / window_size))
    x = windows.view(
        B, H // window_size, W // window_size, window_size, window_size, -1
    )
    x = x.permute(0, 1, 3, 2, 4, 5).contiguous().view(B, H, W, -1)
    return x


class WindowAttention(nn.Module):
    r"""Window based multi-head self attention (W-MSA) module with relative position bias.
    It supports both of shifted and non-shifted window.

    Args:
        dim (int): Number of input channels.
        window_size (tuple[int]): The height and width of the window.
        num_heads (int): Number of attention heads.
        qkv_bias (bool, optional):  If True, add a learnable bias to query, key, value. Default: True
        qk_scale (float | None, optional): Override default qk scale of head_dim ** -0.5 if set
        attn_drop (float, optional): Dropout ratio of attention weight. Default: 0.0
        proj_drop (float, optional): Dropout ratio of output. Default: 0.0
    """

    def __init__(
        self,
        dim,
        window_size,
        num_heads,
        qkv_bias=True,
        qk_scale=None,
        attn_drop=0.0,
        proj_drop=0.0,
    ):

        super().__init__()
        self.dim = dim
        self.window_size = window_size  # Wh, Ww
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = qk_scale or head_dim**-0.5

        # define a parameter table of relative position bias
        self.relative_position_bias_table = nn.Parameter(
            torch.zeros((2 * window_size[0] - 1) * (2 * window_size[1] - 1), num_heads)
        )  # 2*Wh-1 * 2*Ww-1, nH

        # get pair-wise relative position index for each token inside the window
        coords_h = torch.arange(self.window_size[0])
        coords_w = torch.arange(self.window_size[1])
        coords = torch.stack(torch.meshgrid([coords_h, coords_w]))  # 2, Wh, Ww
        coords_flatten = torch.flatten(coords, 1)  # 2, Wh*Ww
        relative_coords = (
            coords_flatten[:, :, None] - coords_flatten[:, None, :]
        )  # 2, Wh*Ww, Wh*Ww
        relative_coords = relative_coords.permute(
            1, 2, 0
        ).contiguous()  # Wh*Ww, Wh*Ww, 2
        relative_coords[:, :, 0] += self.window_size[0] - 1  # shift to start from 0
        relative_coords[:, :, 1] += self.window_size[1] - 1
        relative_coords[:, :, 0] *= 2 * self.window_size[1] - 1
        relative_position_index = relative_coords.sum(-1)  # Wh*Ww, Wh*Ww
        self.register_buffer("relative_position_index", relative_position_index)

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

        trunc_normal_(self.relative_position_bias_table, std=0.02)
        self.softmax = nn.Softmax(dim=-1)

    def forward(self, x, mask=None):
        """
        Args:
            x: input features with shape of (num_windows*B, N, C)
            mask: (0/-inf) mask with shape of (num_windows, Wh*Ww, Wh*Ww) or None
        """
        B_, N, C = x.shape
        qkv = (
            self.qkv(x)
            .reshape(B_, N, 3, self.num_heads, C // self.num_heads)
            .permute(2, 0, 3, 1, 4)
        )
        q, k, v = (
            qkv[0],
            qkv[1],
            qkv[2],
        )  # make torchscript happy (cannot use tensor as tuple)

        q = q * self.scale
        attn = q @ k.transpose(-2, -1)

        relative_position_bias = self.relative_position_bias_table[
            self.relative_position_index.view(-1)
        ].view(
            self.window_size[0] * self.window_size[1],
            self.window_size[0] * self.window_size[1],
            -1,
        )  # Wh*Ww,Wh*Ww,nH
        relative_position_bias = relative_position_bias.permute(
            2, 0, 1
        ).contiguous()  # nH, Wh*Ww, Wh*Ww
        attn = attn + relative_position_bias.unsqueeze(0)

        if mask is not None:
            nW = mask.shape[0]
            attn = attn.view(B_ // nW, nW, self.num_heads, N, N) + mask.unsqueeze(
                1
            ).unsqueeze(0)
            attn = attn.view(-1, self.num_heads, N, N)
            attn = self.softmax(attn)
        else:
            attn = self.softmax(attn)

        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B_, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x

    def extra_repr(self) -> str:
        return f"dim={self.dim}, window_size={self.window_size}, num_heads={self.num_heads}"

    def flops(self, N):
        # calculate flops for 1 window with token length of N
        flops = 0
        # qkv = self.qkv(x)
        flops += N * self.dim * 3 * self.dim
        # attn = (q @ k.transpose(-2, -1))
        flops += self.num_heads * N * (self.dim // self.num_heads) * N
        #  x = (attn @ v)
        flops += self.num_heads * N * N * (self.dim // self.num_heads)
        # x = self.proj(x)
        flops += N * self.dim * self.dim
        return flops


class SwinTransformerBlock(nn.Module):
    r"""Swin Transformer Block.

    Args:
        dim (int): Number of input channels.
        input_resolution (tuple[int]): Input resulotion.
        num_heads (int): Number of attention heads.
        window_size (int): Window size.
        shift_size (int): Shift size for SW-MSA.
        mlp_ratio (float): Ratio of mlp hidden dim to embedding dim.
        qkv_bias (bool, optional): If True, add a learnable bias to query, key, value. Default: True
        qk_scale (float | None, optional): Override default qk scale of head_dim ** -0.5 if set.
        drop (float, optional): Dropout rate. Default: 0.0
        attn_drop (float, optional): Attention dropout rate. Default: 0.0
        drop_path (float, optional): Stochastic depth rate. Default: 0.0
        act_layer (nn.Module, optional): Activation layer. Default: nn.GELU
        norm_layer (nn.Module, optional): Normalization layer.  Default: nn.LayerNorm
    """

    def __init__(
        self,
        dim,
        input_resolution,
        num_heads,
        window_size=7,
        shift_size=0,
        mlp_ratio=4.0,
        qkv_bias=True,
        qk_scale=None,
        drop=0.0,
        attn_drop=0.0,
        drop_path=0.0,
        act_layer=nn.GELU,
        norm_layer=nn.LayerNorm,
    ):
        super().__init__()
        self.dim = dim
        self.input_resolution = input_resolution
        self.num_heads = num_heads
        self.window_size = window_size
        self.shift_size = shift_size
        self.mlp_ratio = mlp_ratio
        if min(self.input_resolution) <= self.window_size:
            # if window size is larger than input resolution, we don't partition windows
            self.shift_size = 0
            self.window_size = min(self.input_resolution)
        assert (
            0 <= self.shift_size < self.window_size
        ), "shift_size must in 0-window_size"

        self.norm1 = norm_layer(dim)
        self.attn = WindowAttention(
            dim,
            window_size=to_2tuple(self.window_size),
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            qk_scale=qk_scale,
            attn_drop=attn_drop,
            proj_drop=drop,
        )

        self.drop_path = DropPath(drop_path) if drop_path > 0.0 else nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(
            in_features=dim,
            hidden_features=mlp_hidden_dim,
            act_layer=act_layer,
            drop=drop,
        )

        if self.shift_size > 0:
            # calculate attention mask for SW-MSA
            H, W = self.input_resolution
            img_mask = torch.zeros((1, H, W, 1))  # 1 H W 1
            h_slices = (
                slice(0, -self.window_size),
                slice(-self.window_size, -self.shift_size),
                slice(-self.shift_size, None),
            )
            w_slices = (
                slice(0, -self.window_size),
                slice(-self.window_size, -self.shift_size),
                slice(-self.shift_size, None),
            )
            cnt = 0
            for h in h_slices:
                for w in w_slices:
                    img_mask[:, h, w, :] = cnt
                    cnt += 1

            mask_windows = window_partition(
                img_mask, self.window_size
            )  # nW, window_size, window_size, 1
            mask_windows = mask_windows.view(-1, self.window_size * self.window_size)
            attn_mask = mask_windows.unsqueeze(1) - mask_windows.unsqueeze(2)
            attn_mask = attn_mask.masked_fill(
                attn_mask != 0, float(-100.0)
            ).masked_fill(attn_mask == 0, float(0.0))
        else:
            attn_mask = None

        self.register_buffer("attn_mask", attn_mask)

    def forward(self, x):
        H, W = self.input_resolution
        B, L, C = x.shape
        assert L == H * W, "input feature has wrong size"

        shortcut = x
        x = self.norm1(x)
        x = x.view(B, H, W, C)

        # cyclic shift
        if self.shift_size > 0:
            shifted_x = torch.roll(
                x, shifts=(-self.shift_size, -self.shift_size), dims=(1, 2)
            )
        else:
            shifted_x = x

        # partition windows
        x_windows = window_partition(
            shifted_x, self.window_size
        )  # nW*B, window_size, window_size, C
        x_windows = x_windows.view(
            -1, self.window_size * self.window_size, C
        )  # nW*B, window_size*window_size, C

        # W-MSA/SW-MSA
        attn_windows = self.attn(
            x_windows, mask=self.attn_mask
        )  # nW*B, window_size*window_size, C

        # merge windows
        attn_windows = attn_windows.view(-1, self.window_size, self.window_size, C)
        shifted_x = window_reverse(attn_windows, self.window_size, H, W)  # B H' W' C

        # reverse cyclic shift
        if self.shift_size > 0:
            x = torch.roll(
                shifted_x, shifts=(self.shift_size, self.shift_size), dims=(1, 2)
            )
        else:
            x = shifted_x
        x = x.view(B, H * W, C)

        # FFN
        x = shortcut + self.drop_path(x)
        x = x + self.drop_path(self.mlp(self.norm2(x)))

        return x

    def extra_repr(self) -> str:
        return (
            f"dim={self.dim}, input_resolution={self.input_resolution}, num_heads={self.num_heads}, "
            f"window_size={self.window_size}, shift_size={self.shift_size}, mlp_ratio={self.mlp_ratio}"
        )

    def flops(self):
        flops = 0
        H, W = self.input_resolution
        # norm1
        flops += self.dim * H * W
        # W-MSA/SW-MSA
        nW = H * W / self.window_size / self.window_size
        flops += nW * self.attn.flops(self.window_size * self.window_size)
        # mlp
        flops += 2 * H * W * self.dim * self.dim * self.mlp_ratio
        # norm2
        flops += self.dim * H * W
        return flops


class PatchMerging(nn.Module):
    r"""Patch Merging Layer.

    Args:
        input_resolution (tuple[int]): Resolution of input feature.
        dim (int): Number of input channels.
        norm_layer (nn.Module, optional): Normalization layer.  Default: nn.LayerNorm
    """

    def __init__(self, input_resolution, dim, norm_layer=nn.LayerNorm):
        super().__init__()
        self.input_resolution = input_resolution
        self.dim = dim
        self.reduction = nn.Linear(4 * dim, 2 * dim, bias=False)
        self.norm = norm_layer(4 * dim)

    def forward(self, x):
        """
        x: B, H*W, C
        """
        H, W = self.input_resolution
        B, L, C = x.shape
        assert L == H * W, "input feature has wrong size"
        assert H % 2 == 0 and W % 2 == 0, f"x size ({H}*{W}) are not even."

        x = x.view(B, H, W, C)

        x0 = x[:, 0::2, 0::2, :]  # B H/2 W/2 C
        x1 = x[:, 1::2, 0::2, :]  # B H/2 W/2 C
        x2 = x[:, 0::2, 1::2, :]  # B H/2 W/2 C
        x3 = x[:, 1::2, 1::2, :]  # B H/2 W/2 C
        x = torch.cat([x0, x1, x2, x3], -1)  # B H/2 W/2 4*C
        x = x.view(B, -1, 4 * C)  # B H/2*W/2 4*C

        x = self.norm(x)
        x = self.reduction(x)

        return x

    def extra_repr(self) -> str:
        return f"input_resolution={self.input_resolution}, dim={self.dim}"

    def flops(self):
        H, W = self.input_resolution
        flops = H * W * self.dim
        flops += (H // 2) * (W // 2) * 4 * self.dim * 2 * self.dim
        return flops


class BasicLayer(nn.Module):
    """A basic Swin Transformer layer for one stage.

    Args:
        dim (int): Number of input channels.
        input_resolution (tuple[int]): Input resolution.
        depth (int): Number of blocks.
        num_heads (int): Number of attention heads.
        window_size (int): Local window size.
        mlp_ratio (float): Ratio of mlp hidden dim to embedding dim.
        qkv_bias (bool, optional): If True, add a learnable bias to query, key, value. Default: True
        qk_scale (float | None, optional): Override default qk scale of head_dim ** -0.5 if set.
        drop (float, optional): Dropout rate. Default: 0.0
        attn_drop (float, optional): Attention dropout rate. Default: 0.0
        drop_path (float | tuple[float], optional): Stochastic depth rate. Default: 0.0
        norm_layer (nn.Module, optional): Normalization layer. Default: nn.LayerNorm
        downsample (nn.Module | None, optional): Downsample layer at the end of the layer. Default: None
        use_checkpoint (bool): Whether to use checkpointing to save memory. Default: False.
    """

    def __init__(
        self,
        dim,
        input_resolution,
        depth,
        num_heads,
        window_size,
        mlp_ratio=4.0,
        qkv_bias=True,
        qk_scale=None,
        drop=0.0,
        attn_drop=0.0,
        drop_path=0.0,
        norm_layer=nn.LayerNorm,
        downsample=None,
        use_checkpoint=False,
    ):

        super().__init__()
        self.dim = dim
        self.input_resolution = input_resolution
        self.depth = depth
        self.use_checkpoint = use_checkpoint

        # build blocks
        self.blocks = nn.ModuleList(
            [
                SwinTransformerBlock(
                    dim=dim,
                    input_resolution=input_resolution,
                    num_heads=num_heads,
                    window_size=window_size,
                    shift_size=0 if (i % 2 == 0) else window_size // 2,
                    mlp_ratio=mlp_ratio,
                    qkv_bias=qkv_bias,
                    qk_scale=qk_scale,
                    drop=drop,
                    attn_drop=attn_drop,
                    drop_path=drop_path[i]
                    if isinstance(drop_path, list)
                    else drop_path,
                    norm_layer=norm_layer,
                )
                for i in range(depth)
            ]
        )

        # patch merging layer
        if downsample is not None:
            self.downsample = downsample(
                input_resolution, dim=dim, norm_layer=norm_layer
            )
        else:
            self.downsample = None

    def forward(self, x):
        for blk in self.blocks:
            if self.use_checkpoint:
                x = checkpoint.checkpoint(blk, x)
            else:
                x = blk(x)
        if self.downsample is not None:
            x = self.downsample(x)
        return x

    def extra_repr(self) -> str:
        return f"dim={self.dim}, input_resolution={self.input_resolution}, depth={self.depth}"

    def flops(self):
        flops = 0
        for blk in self.blocks:
            flops += blk.flops()
        if self.downsample is not None:
            flops += self.downsample.flops()
        return flops


class PatchEmbed(nn.Module):
    r"""Image to Patch Embedding

    Args:
        img_size (int): Image size.  Default: 224.
        patch_size (int): Patch token size. Default: 4.
        in_chans (int): Number of input image channels. Default: 3.
        embed_dim (int): Number of linear projection output channels. Default: 96.
        norm_layer (nn.Module, optional): Normalization layer. Default: None
    """

    def __init__(
        self, img_size=224, patch_size=4, in_chans=3, embed_dim=96, norm_layer=None
    ):
        super().__init__()
        img_size = to_2tuple(img_size)
        patch_size = to_2tuple(patch_size)
        patches_resolution = [
            img_size[0] // patch_size[0],
            img_size[1] // patch_size[1],
        ]
        self.img_size = img_size
        self.patch_size = patch_size
        self.patches_resolution = patches_resolution
        self.num_patches = patches_resolution[0] * patches_resolution[1]

        self.in_chans = in_chans
        self.embed_dim = embed_dim

        self.proj = nn.Conv2d(
            in_chans, embed_dim, kernel_size=patch_size, stride=patch_size
        )
        if norm_layer is not None:
            self.norm = norm_layer(embed_dim)
        else:
            self.norm = None

    def forward(self, x):
        B, C, H, W = x.shape
        # FIXME look at relaxing size constraints
        assert (
            H == self.img_size[0] and W == self.img_size[1]
        ), f"Input image size ({H}*{W}) doesn't match model ({self.img_size[0]}*{self.img_size[1]})."
        x = self.proj(x).flatten(2).transpose(1, 2)  # B Ph*Pw C
        if self.norm is not None:
            x = self.norm(x)
        return x

    def flops(self):
        Ho, Wo = self.patches_resolution
        flops = (
            Ho
            * Wo
            * self.embed_dim
            * self.in_chans
            * (self.patch_size[0] * self.patch_size[1])
        )
        if self.norm is not None:
            flops += Ho * Wo * self.embed_dim
        return flops


class SwinTransformer(nn.Module):
    r"""Swin Transformer
        A PyTorch impl of : `Swin Transformer: Hierarchical Vision Transformer using Shifted Windows`  -
          https://arxiv.org/pdf/2103.14030

    Args:
        img_size (int | tuple(int)): Input image size. Default 224
        patch_size (int | tuple(int)): Patch size. Default: 4
        in_chans (int): Number of input image channels. Default: 3
        num_classes (int): Number of classes for classification head. Default: 1000
        embed_dim (int): Patch embedding dimension. Default: 96
        depths (tuple(int)): Depth of each Swin Transformer layer.
        num_heads (tuple(int)): Number of attention heads in different layers.
        window_size (int): Window size. Default: 7
        mlp_ratio (float): Ratio of mlp hidden dim to embedding dim. Default: 4
        qkv_bias (bool): If True, add a learnable bias to query, key, value. Default: True
        qk_scale (float): Override default qk scale of head_dim ** -0.5 if set. Default: None
        drop_rate (float): Dropout rate. Default: 0
        attn_drop_rate (float): Attention dropout rate. Default: 0
        drop_path_rate (float): Stochastic depth rate. Default: 0.1
        norm_layer (nn.Module): Normalization layer. Default: nn.LayerNorm.
        ape (bool): If True, add absolute position embedding to the patch embedding. Default: False
        patch_norm (bool): If True, add normalization after patch embedding. Default: True
        use_checkpoint (bool): Whether to use checkpointing to save memory. Default: False
    """

    def __init__(
        self,
        img_size=224,
        patch_size=4,
        in_chans=3,
        num_classes=1000,
        embed_dim=128,
        depths=[2, 2, 18, 2],
        num_heads=[4, 8, 16, 32],
        window_size=7,
        mlp_ratio=4.0,
        qkv_bias=True,
        qk_scale=None,
        drop_rate=0.0,
        attn_drop_rate=0.0,
        drop_path_rate=0.5,
        norm_layer=nn.LayerNorm,
        ape=False,
        patch_norm=True,
        use_checkpoint=False,
        **kwargs,
    ):
        super().__init__()

        self.num_classes = num_classes
        self.num_layers = len(depths)
        self.embed_dim = embed_dim
        self.ape = ape
        self.patch_norm = patch_norm
        self.num_features = int(embed_dim * 2 ** (self.num_layers - 1))
        self.mlp_ratio = mlp_ratio

        # split image into non-overlapping patches
        self.patch_embed = PatchEmbed(
            img_size=img_size,
            patch_size=patch_size,
            in_chans=in_chans,
            embed_dim=embed_dim,
            norm_layer=norm_layer if self.patch_norm else None,
        )
        num_patches = self.patch_embed.num_patches
        patches_resolution = self.patch_embed.patches_resolution
        self.patches_resolution = patches_resolution

        # absolute position embedding
        if self.ape:
            self.absolute_pos_embed = nn.Parameter(
                torch.zeros(1, num_patches, embed_dim)
            )
            trunc_normal_(self.absolute_pos_embed, std=0.02)

        self.pos_drop = nn.Dropout(p=drop_rate)

        # stochastic depth
        dpr = [
            x.item() for x in torch.linspace(0, drop_path_rate, sum(depths))
        ]  # stochastic depth decay rule

        # build layers
        self.layers = nn.ModuleList()
        for i_layer in range(self.num_layers - 1):
            layer = BasicLayer(
                dim=int(embed_dim * 2**i_layer),
                input_resolution=(
                    patches_resolution[0] // (2**i_layer),
                    patches_resolution[1] // (2**i_layer),
                ),
                depth=depths[i_layer],
                num_heads=num_heads[i_layer],
                window_size=window_size,
                mlp_ratio=self.mlp_ratio,
                qkv_bias=qkv_bias,
                qk_scale=qk_scale,
                drop=drop_rate,
                attn_drop=attn_drop_rate,
                drop_path=dpr[sum(depths[:i_layer]) : sum(depths[: i_layer + 1])],
                norm_layer=norm_layer,
                downsample=PatchMerging if (i_layer < self.num_layers - 2) else None,
                use_checkpoint=use_checkpoint,
            )
            self.layers.append(layer)

        self.norm = norm_layer(self.num_features)
        self.avgpool = nn.AdaptiveAvgPool1d(1)

        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=0.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)

    @torch.jit.ignore
    def no_weight_decay(self):
        return {"absolute_pos_embed"}

    @torch.jit.ignore
    def no_weight_decay_keywords(self):
        return {"relative_position_bias_table"}

    def forward_features(self, x):
        b, c, h, w = x.shape
        x = self.patch_embed(x)

        if self.ape:
            x = x + self.absolute_pos_embed
        x = self.pos_drop(x)
        s = []
        s.append(
            x.view(b, int((x.shape[1]) ** 0.5), int((x.shape[1]) ** (0.5)), -1)
            .permute(0, 3, 1, 2)
            .contiguous()
        )
        for layer in self.layers:
            x = layer(x)
            s.append(
                x.view(b, int((x.shape[1]) ** 0.5), int((x.shape[1]) ** (0.5)), -1)
                .permute(0, 3, 1, 2)
                .contiguous()
            )

        return s

    def forward(self, x):
        x = self.forward_features(x)
        return x

    def flops(self):
        flops = 0
        flops += self.patch_embed.flops()
        for i, layer in enumerate(self.layers):
            flops += layer.flops()
        return flops



""" U2-Net: Going Deeper with Nested U-Structure for Salient Object Detection

Model from official source: https://github.com/xuebinqin/U-2-Net

No modifications from original implementation
"""
# --------------------------------------------------------
# U2-Net: Going Deeper with Nested U-Structure for Salient Object Detection (https://arxiv.org/pdf/2005.09007.pdf)
# Github source: https://github.com/xuebinqin/U-2-Net
# Apache-2.0 license
# --------------------------------------------------------'
import math

import torch
import torch.nn as nn

__all__ = ["U2NET_full", "U2NET_lite"]


def _upsample_like(x, size):
    return nn.Upsample(size=size, mode="bilinear", align_corners=False)(x)


def _size_map(x, height):
    # {height: size} for Upsample
    size = list(x.shape[-2:])
    sizes = {}
    for h in range(1, height):
        sizes[h] = size
        size = [math.ceil(w / 2) for w in size]
    return sizes


class REBNCONV(nn.Module):
    def __init__(self, in_ch=3, out_ch=3, dilate=1):
        super(REBNCONV, self).__init__()

        self.conv_s1 = nn.Conv2d(
            in_ch, out_ch, 3, padding=1 * dilate, dilation=1 * dilate
        )
        self.bn_s1 = nn.BatchNorm2d(out_ch)
        self.relu_s1 = nn.ReLU(inplace=True)

    def forward(self, x):
        return self.relu_s1(self.bn_s1(self.conv_s1(x)))


class RSU(nn.Module):
    def __init__(self, name, height, in_ch, mid_ch, out_ch, dilated=False):
        super(RSU, self).__init__()
        self.name = name
        self.height = height
        self.dilated = dilated
        self._make_layers(height, in_ch, mid_ch, out_ch, dilated)

    def forward(self, x):
        sizes = _size_map(x, self.height)
        x = self.rebnconvin(x)

        # U-Net like symmetric encoder-decoder structure
        def unet(x, height=1):
            if height < self.height:
                x1 = getattr(self, f"rebnconv{height}")(x)
                if not self.dilated and height < self.height - 1:
                    x2 = unet(getattr(self, "downsample")(x1), height + 1)
                else:
                    x2 = unet(x1, height + 1)

                x = getattr(self, f"rebnconv{height}d")(torch.cat((x2, x1), 1))
                return (
                    _upsample_like(x, sizes[height - 1])
                    if not self.dilated and height > 1
                    else x
                )
            else:
                return getattr(self, f"rebnconv{height}")(x)

        return x + unet(x)

    def _make_layers(self, height, in_ch, mid_ch, out_ch, dilated=False):
        self.add_module("rebnconvin", REBNCONV(in_ch, out_ch))
        self.add_module("downsample", nn.MaxPool2d(2, stride=2, ceil_mode=True))

        self.add_module(f"rebnconv1", REBNCONV(out_ch, mid_ch))
        self.add_module(f"rebnconv1d", REBNCONV(mid_ch * 2, out_ch))

        for i in range(2, height):
            dilate = 1 if not dilated else 2 ** (i - 1)
            self.add_module(f"rebnconv{i}", REBNCONV(mid_ch, mid_ch, dilate=dilate))
            self.add_module(
                f"rebnconv{i}d", REBNCONV(mid_ch * 2, mid_ch, dilate=dilate)
            )

        dilate = 2 if not dilated else 2 ** (height - 1)
        self.add_module(f"rebnconv{height}", REBNCONV(mid_ch, mid_ch, dilate=dilate))


class U2NET(nn.Module):
    def __init__(self, cfgs, out_ch):
        super(U2NET, self).__init__()
        self.out_ch = out_ch
        self._make_layers(cfgs)

    def forward(self, x):
        sizes = _size_map(x, self.height)
        maps = []  # storage for maps

        # side saliency map
        def unet(x, height=1):
            if height < 6:
                x1 = getattr(self, f"stage{height}")(x)
                x2 = unet(getattr(self, "downsample")(x1), height + 1)
                x = getattr(self, f"stage{height}d")(torch.cat((x2, x1), 1))
                side(x, height)
                return _upsample_like(x, sizes[height - 1]) if height > 1 else x
            else:
                x = getattr(self, f"stage{height}")(x)
                side(x, height)
                return _upsample_like(x, sizes[height - 1])

        def side(x, h):
            # side output saliency map (before sigmoid)
            x = getattr(self, f"side{h}")(x)
            x = _upsample_like(x, sizes[1])
            maps.append(x)

        def fuse():
            # fuse saliency probability maps
            maps.reverse()
            x = torch.cat(maps, 1)
            x = getattr(self, "outconv")(x)
            maps.insert(0, x)
            return [torch.sigmoid(x) for x in maps]

        unet(x)
        maps = fuse()
        return maps

    def _make_layers(self, cfgs):
        self.height = int((len(cfgs) + 1) / 2)
        self.add_module("downsample", nn.MaxPool2d(2, stride=2, ceil_mode=True))
        for k, v in cfgs.items():
            # build rsu block
            self.add_module(k, RSU(v[0], *v[1]))
            if v[2] > 0:
                # build side layer
                self.add_module(
                    f"side{v[0][-1]}", nn.Conv2d(v[2], self.out_ch, 3, padding=1)
                )
        # build fuse layer
        self.add_module(
            "outconv", nn.Conv2d(int(self.height * self.out_ch), self.out_ch, 1)
        )


def U2NET_full(**kwargs):
    full = {
        # cfgs for building RSUs and sides
        # {stage : [name, (height(L), in_ch, mid_ch, out_ch, dilated), side]}
        "stage1": ["En_1", (7, 3, 32, 64), -1],
        "stage2": ["En_2", (6, 64, 32, 128), -1],
        "stage3": ["En_3", (5, 128, 64, 256), -1],
        "stage4": ["En_4", (4, 256, 128, 512), -1],
        "stage5": ["En_5", (4, 512, 256, 512, True), -1],
        "stage6": ["En_6", (4, 512, 256, 512, True), 512],
        "stage5d": ["De_5", (4, 1024, 256, 512, True), 512],
        "stage4d": ["De_4", (4, 1024, 128, 256), 256],
        "stage3d": ["De_3", (5, 512, 64, 128), 128],
        "stage2d": ["De_2", (6, 256, 32, 64), 64],
        "stage1d": ["De_1", (7, 128, 16, 64), 64],
    }
    return U2NET(cfgs=full, out_ch=1)


def U2NET_lite(**kwargs):
    lite = {
        # cfgs for building RSUs and sides
        # {stage : [name, (height(L), in_ch, mid_ch, out_ch, dilated), side]}
        "stage1": ["En_1", (7, 3, 16, 64), -1],
        "stage2": ["En_2", (6, 64, 16, 64), -1],
        "stage3": ["En_3", (5, 64, 16, 64), -1],
        "stage4": ["En_4", (4, 64, 16, 64), -1],
        "stage5": ["En_5", (4, 64, 16, 64, True), -1],
        "stage6": ["En_6", (4, 64, 16, 64, True), 64],
        "stage5d": ["De_5", (4, 128, 16, 64, True), 64],
        "stage4d": ["De_4", (4, 128, 16, 64), 64],
        "stage3d": ["De_3", (5, 128, 16, 64), 64],
        "stage2d": ["De_2", (6, 128, 16, 64), 64],
        "stage1d": ["De_1", (7, 128, 16, 64), 64],
    }
    return U2NET(cfgs=lite, out_ch=1)



import os
from pathlib import Path

import cv2
import numpy as np


def read_rgb(file_path: Path) -> np.array:
    """Load an image from file_path as a numpy array

    Args:
        file_path (Path): path to image

    Raises:
        FileNotFoundError: if file is not found

    Returns:
        np.array: image
    """
    if not os.path.exists(file_path):
        raise FileNotFoundError(f"The path {file_path} does not exist")
    image = cv2.imread(file_path)
    if image is None:
        raise ValueError(f"Unable to read {file_path}.")
    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
    return image


def read_mask(file_path: Path) -> np.array:
    """Load a segmentation mask from file_path as a numpy array

    Args:
        file_path (Path): path to image

    Raises:
        FileNotFoundError: if file is not found

    Returns:
        np.array: mask
    """
    if not os.path.exists(file_path):
        raise FileNotFoundError(f"The path {file_path} does not exist")
    img = cv2.imread(str(file_path), cv2.IMREAD_GRAYSCALE)
    if img is None:
        raise ValueError(f"Unable to read {file_path}.")

    img = img.astype("float64")
    if img.max() == 255:
        img /= 255
    return img


def save_image(image: np.array, output_path: str):
    """Save an image at given path making sure the folder exists

    Args:
        image (np.array): image to save
        output_path (str): output path
    """
    output_dir = output_path.replace(os.path.basename(output_path), "")
    if output_dir != "":
        os.makedirs(output_dir, exist_ok=True)

    if len(image.shape) > 2:
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
    try:
        cv2.imwrite(output_path, image)
    except Exception as excp:
        print(
            f"[ERROR] While saving image at \
                path {output_path} found an error - {excp}"
        )



import cv2
import matplotlib.pyplot as plt
import numpy as np

from saldet.utils import device


def prepare_mask(
    mask: np.array, width: int, height: int, threshold: float = 0.5
) -> np.array:
    """Prepare mask to save

    Args:
        mask (np.array): mask to save
        width (int): resize width
        height (int): resize height
        threshold (float, optional): mask threshold. Defaults to .5.

    Returns:
        np.array: prepared mask
    """
    if get_device() != "cpu":
        mask = mask.squeeze().cpu().numpy()
    else:
        mask = mask.squeeze().numpy()
    mask = cv2.resize(mask, (width, height))
    if threshold is not None:
        mask[mask >= threshold] = 1
        mask[mask < threshold] = 0
    return mask


def apply_mask(image: np.array, mask: np.array, alpha: float = 1.0) -> np.array:
    """Apply segmentation masks to an image

    Args:
        image (np.array): image to be blended with the mask.
        mask (np.array): binary mask to be blended with the image.
        alpha (float): relative weight of the mask, 1.0 means no transparency, 0.0 means the mask is completely transparent.
            Defaults to 1.0.

    Returns:
        np.array: image with masks applied
    """

    assert np.unique(mask).shape[0] <= 2, "Mask must be binary"
    out = image.copy()
    cmap = plt.cm.tab10
    for cat_id in np.unique(mask)[1:]:
        bool_mask = mask == cat_id
        color = np.array(cmap(cat_id - 1)[:3]) * 255
        cat_mask = (np.expand_dims(mask, axis=2) * color)[bool_mask].astype(np.uint8)
        out[bool_mask] = cv2.addWeighted(
            out[bool_mask], 1.0 - alpha, cat_mask, alpha, 0.0
        )

    return out


def plot_predictions(
    image: np.array, gt_mask: np.array, mask: np.array, alpha: float = 1.0
):
    """Plot ground truth and predicted mask on original image in one row.

    Args:
        image (np.array): original image
        gt_mask (np.array): ground truth image (can be None)
        mask (np.array): predicted mask
        alpha (float): relative weight of the mask, 1.0 means no transparency, 0.0 means the mask is completely transparent.
            Defaults to 1.0.
    """

    plt.figure(figsize=(32, 9))
    plt.subplot(1, 3, 1)
    plt.title("Ground Truth")
    if gt_mask is not None:
        gt_image = apply_mask(image=image, mask=gt_mask, alpha=alpha)
    else:
        gt_image = image

    plt.imshow(gt_image)

    plt.subplot(1, 3, 2)
    pred_image = apply_mask(image=image, mask=mask, alpha=alpha)
    plt.title("Model prediction")
    plt.imshow(pred_image)
    plt.show()



import datetime

STRFTIME_FORMAT = "%Y-%m-%d-%H-%M-%S"


def now() -> str:
    """Return current datetime as string format as "%Y-%m-%d-%H-%M-%S"

    Returns:
        str: current datetime (format "%Y-%m-%d-%H-%M-%S")
    """
    STRFTIME_FORMAT = "%Y-%m-%d-%H-%M-%S"
    return datetime.datetime.now().strftime(STRFTIME_FORMAT)



import os

import torch
import torch.nn as nn

from saldet.pl import SaliencyPLModel
from saldet.utils import device

from .models import *

__all__ = ["create_model", "load_checkpoint"]

FACTORY = {
    "u2net_lite": U2NET_lite,
    "u2net_full": U2NET_full,
    "pgnet": PGNet,
    "pfan": PFAN,
}


def create_model(model_name: str, checkpoint_path: str = None, **kwargs) -> nn.Module:
    """Create a saliency model

    Args:
        model_name (str): model name
        checkpoint_path (str, optional): path to model checkpoint. Defaults to None.

    Returns:
        nn.Module: saliency model
    """

    assert model_name in FACTORY.keys(), f"Model {model_name} not supported."
    model = FACTORY[model_name](**kwargs)
    if checkpoint_path is not None:
        assert os.path.exists(checkpoint_path), f"{checkpoint_path} does not exist."
        try:
            state_dict = torch.load(f=checkpoint_path, map_location=device())
        except Exception as e:
            print(f"Found an exception while loading state dict - {e}")
            print(f"Trying on cpu")
            state_dict = torch.load(f=checkpoint_path, map_location="cpu")
        model.load_state_dict(state_dict=state_dict)
        print(f"> Loaded state dict for {model_name}.")

    return model


def list_models():
    """List all available models"""
    print(f"saldet available models:")
    for m in FACTORY:
        print(f"> {m}")


def load_checkpoint(
    ckpt: str,
    model_name: str = None,
) -> nn.Module:
    """Load either a checkpoint with pytorch-lightning support or a pth file

    Args:
        ckpt (str): path to ckpt/pth file
        model_name (str, optional): model name. Defaults to None.

    Returns:
        nn.Module: model
    """
    if ckpt.endswith(".pth"):
        model = create_model(model_name=model_name, checkpoint_path=ckpt)
    elif ckpt.endswith(".ckpt"):
        try:
            model = SaliencyPLModel.load_from_checkpoint(
                checkpoint_path=ckpt, map_location=device()
            )
            if hasattr(model, "criterion") and hasattr(model, "model"):
                model = model.model
            print(f"> Loaded state dict from {ckpt}.")
        except Exception as e:
            print(f"> [ERROR] Not able to load ckpt from {ckpt} -  Exception: {e}")
    else:
        print(f"> [ERROR] {os.path.splitext(ckpt)[-1]} not supported.")
        quit()

    return model



import torch
import torch.nn as nn
from torch.nn import functional as F

from saldet.ops import flat


class AttentionGuidedLoss(nn.Module):
    def __init__(self) -> None:
        super().__init__()

    def forward(
        self, pred: torch.Tensor, mask: torch.Tensor, p4: torch.Tensor, p5: torch.Tensor
    ) -> torch.Tensor:

        g = flat(mask)
        np4 = torch.sigmoid(p4.detach())
        np5 = torch.sigmoid(p5.detach())
        p4 = flat(np4)
        p5 = flat(np5)
        w1 = torch.abs(g - p4)
        w2 = torch.abs(g - p5)
        w = (w1 + w2) * 0.5 + 1
        attn_bce = F.binary_cross_entropy_with_logits(
            pred, g, weight=w * 1.0, reduction="mean"
        )
        return attn_bce



from typing import List

import torch
import torch.nn as nn
from torch.nn import functional as F

from saldet.loss.attn_guided import AttentionGuidedLoss
from saldet.loss.bce_iou_loss import BCEIoULoss


class PGLoss(nn.Module):
    def __init__(self, w: float = 0.125) -> None:
        super().__init__()
        self.attn_loss = AttentionGuidedLoss()
        self.bce_iou_loss = BCEIoULoss()
        self.w = w

    def forward(self, preds: List[torch.Tensor], target: torch.Tensor) -> torch.Tensor:

        p1, wr, ws, attn_map = preds
        attn_loss = self.attn_loss(attn_map, target, wr, ws)  # attention guided loss
        loss1 = self.bce_iou_loss(p1, target)  # loss_b+i
        loss2 = (
            self.bce_iou_loss(wr, target) * self.w
            + self.bce_iou_loss(ws, target) * self.w
        )

        return loss1 + loss2 + attn_loss



from typing import Any, List, Tuple, Union

import pytorch_lightning as pl
import torch
from torch import nn
from torch.nn.modules.loss import _Loss
from torch.optim import Optimizer
from torch.optim.lr_scheduler import _LRScheduler

from saldet.utils import device


class SaliencyPLModel(pl.LightningModule):
    """Saliency Model with PyTorchLightning Module

    Args:
        model (nn.Module): saliency model
        criterion (_Loss): criterion for loss
        optimizer (Optimizer): optimizer
        lr_scheduler (_LRScheduler, optional): Optional; Learning rate scheduler. Defaults to None.
    """

    def __init__(
        self,
        model: nn.Module,
        criterion: _Loss,
        optimizer: Optimizer,
        lr_scheduler: _LRScheduler = None,
    ) -> None:

        super().__init__()
        self.save_hyperparameters()
        self.model = model
        self.criterion = criterion
        self.optimizer = optimizer
        self.lr_scheduler = lr_scheduler

        # TODO: add BinaryMetrics

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # Output a tensor of shape (batch size, num classes, height, width)
        return self.model(x)

    def training_step(self, batch: torch.Tensor, batch_idx: int) -> float:
        """Training step

        Args:
            batch (torch.Tensor): batch of images and masks
            batch_idx (int): batch index

        Returns:
            float: loss value
        """
        x, mask = batch
        preds = self(x)
        # Loss
        loss = self.criterion(preds, mask)
        self.log("loss_train", loss, sync_dist=True)
        self.log("lr", self.lr_scheduler.get_last_lr()[0], prog_bar=True)
        return loss

    def validation_step(self, batch: torch.Tensor, batch_idx: int) -> None:
        """Validation step

        Args:
            batch (torch.Tensor): batch of images and masks
            batch_idx (int): batch index

        """
        x, mask = batch
        preds = self(x)
        loss = self.criterion(preds, mask)

        self.log("loss_val", loss, sync_dist=True, prog_bar=True)

    def configure_optimizers(
        self,
    ) -> Union[List[Optimizer], Tuple[List[Optimizer], List[_LRScheduler]]]:
        if self.lr_scheduler is None:
            return [self.optimizer]
        else:
            return [self.optimizer], [self.lr_scheduler]



from typing import Iterable

from torch.optim import SGD, Adam, AdamW, Optimizer

__all__ = ["create_optimizer"]

FACTORY = {
    "sgd": SGD,
    "adam": Adam,
    "adamw": AdamW,
}


def create_optimizer(name: str, params: Iterable, **kwargs) -> Optimizer:
    """Return an optimizer

    Args:
        name (str): optimizer name
        params (Iterable): optimizer
        split_by (str, optional): splitting parameters into base and head based on str (only for PGNet). Defaults to None.

    Returns:
        Optimizer: optimizer

    """
    name = name.lower()
    assert (
        name in FACTORY.keys()
    ), f"Only {list(FACTORY.keys())} optimizers are supported. Change {name} to one of them."
    return FACTORY[name](params, **kwargs)



from torch.optim import Optimizer
from torch.optim.lr_scheduler import (
    CosineAnnealingLR,
    CosineAnnealingWarmRestarts,
    LinearLR,
    StepLR,
    _LRScheduler,
)

__all__ = ["create_scheduler"]

FACTORY = {
    "cosine": CosineAnnealingLR,
    "cosine_restarts": CosineAnnealingWarmRestarts,
    "linear": LinearLR,
    "step": StepLR,
}


def create_scheduler(optimizer: Optimizer, name: str, **kwargs) -> _LRScheduler:
    name = name.lower()
    assert (
        name in FACTORY.keys()
    ), f"Only {list(FACTORY.keys())} lr_schedulers are supported. Change {name} to one of them."
    return FACTORY[name](optimizer, **kwargs)
